"""Contains utility functions / classes."""

import contextlib
import os
import re
from typing import Dict, Any, Pattern, Iterable, Callable, List, Tuple, Iterator, Optional, cast

from .types import YamlDocument
from .validator import Validator

PatternGenerator = Iterator[Tuple[Dict[Any, Any], str, Any]]
PathEvalCallable = Callable[[str, Optional[str]], Optional[str]]


def dict_find_pattern(_dict: Dict[Any, Any], pattern: str, search_in_keys: bool = True,
                      search_in_values: bool = True) -> PatternGenerator:
    """
    Find keys and values in a (nested) dictionary with regular expression pattern.
    Examples:
        >>> from collections import OrderedDict
        >>> d = OrderedDict((
        ...     ('foo', 'bar'),
        ...     ('bar', 'baz'),
        ...     ('nested', dict(p1=1, p2=2)),
        ...     ('lst', ['bar', 'foo', dict(p3=3, p4=4)])
        ... ))
        >>> f = dict_find_pattern(d, '.*a.*', search_in_keys=True, search_in_values=False)
        >>> def print_findings(f):
        ...     for _, key, value in f:
        ...         print(key, value)
        >>> print_findings(f)
        bar baz
        >>> f = dict_find_pattern(d, '.*a.*', search_in_keys=False, search_in_values=True)
        >>> print_findings(f)
        foo bar
        bar baz

    Args:
        _dict: Dictionary to scan for the specified pattern.
        pattern: The pattern to scan for.
        search_in_keys: If True keys will be probed for the pattern; otherwise not.
        search_in_values: If True values will be probed for the pattern; otherwise not.

    Returns:
        Generator to iterate over the findings.
    """

    def find(dic: Dict[Any, Any], regex: Pattern[str]) -> PatternGenerator:
        Validator.instance_of(target_type=dict, raise_ex=True, dic=dic)

        for k, val in dic.items():
            if search_in_keys and isinstance(k, str) and regex.match(k):
                yield dic, k, val
            if search_in_values and isinstance(val, str) and regex.match(val):
                yield dic, k, val
            if isinstance(val, list):
                for litem in val:
                    if isinstance(litem, dict):
                        for j in find(litem, regex):
                            yield j
            if isinstance(val, dict):
                for j in find(val, regex):
                    yield j

    regex = re.compile(pattern)
    return find(_dict, regex)


@contextlib.contextmanager
def modified_environ(*remove: str, **update: str) -> Iterator[None]:
    """
    Temporarily updates the ``os.environ`` dictionary in-place and resets it to the original state
    when finished.
    (https://stackoverflow.com/questions/2059482/
        python-temporarily-modify-the-current-processs-environment/34333710#34333710)

    The ``os.environ`` dictionary is updated in-place so that the modification is sure to work in
    all situations.

    Args:
        remove: Environment variables to remove.
        update: Dictionary of environment variables and values to add/update.

    Examples:
        >>> with modified_environ(Test='abc'):
        ...     import os
        ...     print(os.environ.get('Test'))
        abc
        >>> print(os.environ.get('Test'))
        None
    """
    env = os.environ
    update = update or {}
    remove = remove or ()

    # List of environment variables being updated or removed.
    stomped = (set(update.keys()) | set(remove)) & set(env.keys())
    # Environment variables and values to restore on exit.
    update_after = {k: env[k] for k in stomped}
    # Environment variables and values to remove on exit.
    remove_after = frozenset(k for k in update if k not in env)

    try:
        env.update(update)
        [env.pop(k, None) for k in remove]  # pylint: disable=expression-not-assigned
        yield
    finally:
        env.update(update_after)
        [env.pop(k) for k in remove_after]  # pylint: disable=expression-not-assigned


def eval_first_non_none(eval_list: Iterable[Callable[..., Any]], **kwargs: Any) -> Any:
    """
    Executes a list of functions and returns the first non none result. All kwargs will be passed as
    kwargs to each individual function. If all functions return None, None is the overall result.

    Examples:

        >>> eval_first_non_none((lambda: None, lambda: None, lambda: 3))
        3
        >>> print(eval_first_non_none([lambda: None, lambda: None, lambda: None]))
        None
        >>> eval_first_non_none([
        ...     lambda cnt: cnt if cnt == 1 else None,
        ...     lambda cnt: cnt if cnt == 2 else None,
        ...     lambda cnt: cnt if cnt == 3 else None]
        ... , cnt=2)
        2
    """
    Validator.is_real_iterable(raise_ex=True, eval_list=eval_list)

    for eval_fun in eval_list:
        res = eval_fun(**kwargs)
        if res is not None:
            return res
    return None


class FileLocator:  # pylint: disable=too-few-public-methods
    """
    Based on a base_path, the given file path and a possible parent file path the locator
    will construct a absolute file_path if the given file name is relative.

    If the file path is absolute the absolute file path will be used. If it's relative and a
    base_path is given the both will be concatenated. If no base path is given, the path from the
    parent file path will be used. If there is no parent file path the current working directory is
    the default.
    """

    def __init__(self, base_path: Optional[str] = None, parent_overrides_base: bool = False):
        """
        Args:
            base_path:
            parent_overrides_base: If set to True the file path of the parent (if any) will
                overrule the base_path; otherwise the base_path will overrule the parent's file
                path even (if any).
        """
        self.base_path = base_path and str(base_path)
        self.parent_overrides_base = bool(parent_overrides_base)

    @staticmethod
    def _eval_absolute_path(abs_or_rel_file_path: str,
                            parent_file_path: Optional[str] = None) -> Optional[str]:  # pylint: disable=unused-argument
        if os.path.isabs(abs_or_rel_file_path):
            return abs_or_rel_file_path
        return None

    def _eval_base_path(self, abs_or_rel_file_path: str,
                        parent_file_path: Optional[str] = None) -> Optional[str]:  # pylint: disable=unused-argument
        if self.base_path is not None:
            return os.path.join(self.base_path, abs_or_rel_file_path)
        return None

    @staticmethod
    def _eval_parent_file_path(abs_or_rel_file_path: str,
                               parent_file_path: Optional[str]) -> Optional[str]:
        if parent_file_path is not None and Validator.is_file(parent_file_path=parent_file_path):
            # If the yaml data is from a file, we assume the base_path should be the path where
            # the file is located
            return os.path.join(os.path.dirname(parent_file_path), abs_or_rel_file_path)
        return None

    @staticmethod
    def _eval_cwd(abs_or_rel_file_path: str,
                  parent_file_path: Optional[str] = None) -> Optional[str]:  # pylint: disable=unused-argument
        return os.path.join(os.getcwd(), abs_or_rel_file_path)

    def _eval_list(self) -> Iterable[PathEvalCallable]:
        eval_list = [self._eval_absolute_path]  # type: List[PathEvalCallable]
        if self.parent_overrides_base:
            return eval_list + [self._eval_parent_file_path, self._eval_base_path, self._eval_cwd]
        return eval_list + [self._eval_base_path, self._eval_parent_file_path, self._eval_cwd]

    def __call__(self, abs_or_rel_file_path: str,
                 parent_file_path: Optional[YamlDocument] = None) -> str:
        """
        Given a file_path and the actual file path of the parent file the absolute path of the
        potential relative path will be determined. If `parent_file_path` is not None that basically
        means that the file to locate is part of an externally loaded file.

        Args:
            abs_or_rel_file_path: Absolute or relative file path.
            parent_file_path: When it's a file it is used to determine a base path if necessary.

        Returns:
            Returns the absolute path of the file. If it's already absolute, nothing changes.
        """
        return cast(str, eval_first_non_none(
            self._eval_list(),
            abs_or_rel_file_path=abs_or_rel_file_path,
            parent_file_path=parent_file_path
        ))



"""Contains extensions for dictmentor."""

import os
import re
from abc import abstractmethod
from typing import Any, Pattern, Union, cast, Optional, Dict

import attr

from .types import ExtensionConfig, AugmentedDict, YamlDocument, NodeKeyVal
from .utils import FileLocator
from .validator import Validator


class ExtensionError(Exception):
    """Is raised when an error occurrs within an extension."""


@attr.s
class ExtensionContext:  # pylint: disable=too-few-public-methods
    """
    Context for processing
        * mentor: The processor (DictMentor) instance.
        * document: The document as it was passed to the processor. Might be a file name,
            stream, string or dictionary.
        * dct: The whole dictionary not just parts of it.
        * parent_node: The parent_node of node that has the configured pattern as a python
            dictionary.
        * node: Tuple of key and value of the node that matches the pattern.
    """
    # Actual a DictMentor instance, but not explicit due to circle ref
    mentor = attr.ib()  # type: Any
    document = attr.ib()  # type: Union[YamlDocument, AugmentedDict]
    dct = attr.ib()  # type: AugmentedDict
    parent_node = attr.ib()  # type: Dict[Any, Any]
    node = attr.ib()  # type: NodeKeyVal


class Extension:
    """
    Provides a general interface for any extension.
    """
    @classmethod
    def is_valid_extension(cls, candidate: Any) -> bool:
        """Checks if the given candidate is a valid extension for dictmentor."""
        return hasattr(candidate, 'apply') and hasattr(candidate, 'config')

    def config(self) -> ExtensionConfig:
        """
        Instructs the `DictMentor` scanner which nodes are relevant for this extension.
        A config consists of three configurable instruction arguments:
        * pattern -> Specifies the pattern to look for in the dictionary model (standard regex)
        * key -> If set to True, the pattern will be applied to keys; otherwise keys are ignored.
        * value -> If set to True, the pattern will be applied to values; otherwise values are
            ignored.
        Under the hood the config will be passed to dict_find_pattern by the mentor.

        Examples:
            >>> # Looks for the given pattern in key strings only
            >>> cfg = dict(pattern='external.*', search_in_keys=True, search_in_values=False)
            >>> from collections import OrderedDict
            >>> dic = OrderedDict((
            ...     ('a', 'a'),
            ...     ('external12', 'abc'),
            ...     ('b', dict(c='c', external56='cde'))
            ... ))
            >>> from dictmentor.utils import dict_find_pattern
            >>> for _, key, value in dict_find_pattern(dic, **cfg):
            ...     print(key, value)
            external12 abc
            external56 cde

        Returns:
        """
        return self._config()

    @abstractmethod
    def _config(self) -> ExtensionConfig:
        raise NotImplementedError()  # pragma: no cover

    def apply(self, ctx: ExtensionContext) -> AugmentedDict:
        """
        This method gets called every time a configured pattern is found in the dictionary.
        See Args section for the available context that is passed.

        Args:
            ctx (dict): The processing context.

        Returns:
            Returns a python dictionary. The dictionary will be integrated into dictionary model.
            Basically parent_node.update(result) will be called to modify the dictionarx model.
            This way you can modify even more than just the actual key / value pair.
            See the implementation of `ExternalResource` which leverages this feature.
        """
        return self._apply(ctx)

    @abstractmethod
    def _apply(self, ctx: ExtensionContext) -> AugmentedDict:
        raise NotImplementedError()  # pragma: no cover


class ExternalResource(Extension):
    """
    Loads an external resource (text) into the dictionary.
    Syntax: {{external::/path/to/external/file.txt}}

    Examples:
        >>> import tempfile
        >>> tmp = tempfile.NamedTemporaryFile()
        >>> with open(tmp.name, 'w') as fp:
        ...     print('SELECT *', file=fp)
        ...     print('FROM my_table', file=fp)
        ...     print("WHERE 1 = 1", file=fp)
        >>> current = '{"a": 1, "b": 2, "sql": "{{external::%s}}"}' % tmp.name
        >>> from dictmentor import DictMentor
        >>> (DictMentor().bind(ExternalResource()).load_yaml(current)
        ... == {'a': 1, 'b': 2, 'sql': 'SELECT *\\nFROM my_table\\nWHERE 1 = 1\\n'})
        True
    """
    __pattern__ = '.*({{external::(.*)}}).*'

    def __init__(self, base_path: Optional[str] = None, locator: Optional[FileLocator] = None):
        """
        Initializer.

        Args:
            base_path: Base path. Fast way to construct a `FileLocator` if no locator is passed.
                Will be ignored when a locator is given.
            locator:
        """
        self._base_path = str(base_path)
        if locator is None:
            self.locator = FileLocator(self._base_path)
        else:
            self.locator = locator

    def _config(self) -> ExtensionConfig:
        """
        Tells the processor to scan for the given pattern on node values only.
        Returns: The processor scanning configuration.
        """
        return dict(pattern=self.__pattern__, search_in_keys=False, search_in_values=True)

    def _apply(self, ctx: ExtensionContext) -> AugmentedDict:
        """
        Performs the actual loading of an external resource into the current model.

        Args:
            ctx: The processing context.

        Returns:
            Returns a dictionary that gets incorporated into the actual model.
        """
        def process(pattern: Pattern[str], _str: str) -> Any:
            _match = pattern.match(_str)
            if _match is None:
                return _str  # pragma: no cover
            # We got a match
            # Group 0: Whole match; Group 1: Our placeholder;
            # Group 2: file path to external resource
            placeholder, external_path = _match.group(1), _match.group(2)
            with open(self.locator(
                    external_path,
                    cast(str, ctx.document) if Validator.is_file(document=ctx.document) else None
            )) as fhandle:
                # Json does not support line breaks. We will have to mask them
                content = fhandle.read()
            return _str.replace(placeholder, content)

        node_key, node_value = ctx.node
        _pattern = re.compile(self.__pattern__)
        return {node_key: process(_pattern, node_value)}


class ExternalYamlResource(ExternalResource):
    """
    Enables the integration of external/other yaml files into the actual model.

    Examples:
        current.json:
        {
            "a": 1
            "b": 2
            "external": "/path/to/other/json/file.json"
        }
        /path/to/other/json/file.json:
        {
            "b": 99,
            "c": 3
        }
        Will result into:
        {
            "a": 1,
            "b": 99,
            "c": 3
        }
        >>> import tempfile
        >>> tmp = tempfile.NamedTemporaryFile()
        >>> with open(tmp.name, 'w') as fp:
        ...     print('{"b": 99, "c": 3}', file=fp)
        >>> current = '{"a": 1, "b": 2, "external": "%s"}' % tmp.name
        >>> from dictmentor import DictMentor
        >>> (DictMentor().bind(ExternalYamlResource()).load_yaml(current)
        ... == {'a': 1, 'b': 99, 'c': 3})
        True
    """
    __pattern__ = r'^\s*external\s*$'

    def __init__(self, **kwargs: Any):
        """
        Initializer.

        Args:
            base_path: Override the path for external resource lookups with relative links.
                Default path is where the yaml file is located. When the yaml data comes from a
                stream or a str (not a file) the default is the current working directory.
        """
        super().__init__(**kwargs)

    def _config(self) -> ExtensionConfig:
        """
        Tells the processor to look for the specified pattern in node key only.
        Returns: The scanner configuration.
        """
        return dict(pattern=self.__pattern__, search_in_keys=True, search_in_values=False)

    def _apply(self, ctx: ExtensionContext) -> Any:
        """
        Loads a yaml fragment from an external file.

        Args:
            ctx: The processing context.

        Returns:
            The external resource as a python dictionary. The fragment is already send through
            the processor as well.
        """
        _, external_path = ctx.node
        return ctx.mentor.load_yaml(self.locator(
            external_path,
            cast(str, ctx.document) if Validator.is_file(document=ctx.document) else None
        ))


class Environment(Extension):
    """
    Enables the replacement of any environment variables in a dictionary.
    Syntax: {{env::my_env}} -> Will be replaced by the value of the environment variable 'my_env'

    Examples:
        >>> from dictmentor import DictMentor
        >>> from dictmentor.utils import modified_environ
        >>> jstr = '{"a": 1, "file_path": "my_file.{{env::ENVIRONMENT}}.cfg"}'
        >>> with modified_environ(ENVIRONMENT='production'):
        ...     (DictMentor().bind(Environment()).load_yaml(jstr) ==
        ...         {'a': 1, 'file_path': 'my_file.production.cfg'})
        True
        >>> # If an environment variable is not set, a default is rendered instead.
        >>> # You can specify the default via an init argument. Default of the default is 'none' ...
        >>> with modified_environ('ENVIRONMENT'):
        ...     (DictMentor().bind(Environment(default='development')).load_yaml(jstr)
        ...     == {'a': 1, 'file_path': 'my_file.development.cfg'})
        True
        >>> # ... or by specifying fail_on_unset an exception will be raised
        >>> with modified_environ('ENVIRONMENT'):
        ...     res = DictMentor().bind(
        ...         Environment(fail_on_unset=True, default='development')
        ...     ).load_yaml(jstr)
        Traceback (most recent call last):
        ...
        dictmentor.extensions.ExtensionError: Environment variable 'ENVIRONMENT' is unset.
    """
    __pattern__ = '.*({{env::(.*?)(:=(.*))?}}).*'

    def __init__(self, fail_on_unset: bool = False, default: str = 'none'):
        """
        Initializer.
        Args:
            fail_on_unset (bool): If set to True an exception will be raised when the environment
                variable is unset; otherwise the default value (see next) will be used instead.
            default (str): If a environment variable is unset, it will get this value instead.
        """
        self.fail_on_unset = bool(fail_on_unset)
        self.default = str(default)

    def _config(self) -> ExtensionConfig:
        """
        Tells the processor to scan for the specified pattern in node keys and values.
        Returns: The scanner configuration.
        """
        # tells the scanner to look for the specified pattern in key and value strings
        return dict(pattern=self.__pattern__, search_in_keys=True, search_in_values=True)

    def _apply(self, ctx: ExtensionContext) -> AugmentedDict:
        """
        Replaces any {{env::*}} directives with it's actual environment variable value or a default.

        Args:
            ctx: The processing context.

        Returns:
            Returns the altered node key and value.
        """
        node_key, node_value = ctx.node

        def process(pattern: Pattern[str], _str: str) -> str:
            _match = pattern.match(_str)
            if _match is None:
                return _str
            # We got a match
            # Group 0: Whole match; Group 1: Our placeholder; Group 2: The environment variable
            placeholder, envvar, default = _match.group(1), _match.group(2), _match.group(4)
            envvalue = os.environ.get(envvar, None)
            if envvalue is None and default is None and self.fail_on_unset:
                raise ExtensionError("Environment variable '{}' is unset.".format(envvar))

            return _str.replace(placeholder, envvalue or default or self.default)

        _pattern = re.compile(self.__pattern__)
        node_key = process(_pattern, node_key)
        node_value = process(_pattern, node_value)

        return {node_key: node_value}


class Variables(Extension):
    """
    Enables the replacement of any variables in a dictionary.
    Syntax: {{var::my_env}} -> Will be replaced by the value of the variable 'my_env'

    Examples:
        >>> from dictmentor import DictMentor
        >>> jstr = '{"a": 1, "file_path": "my_file.{{var::environment}}.cfg"}'
        >>> extension = Variables(environment='production')
        >>> (DictMentor().bind(extension).load_yaml(jstr) ==
        ...     {'a': 1, 'file_path': 'my_file.production.cfg'})
        True
        >>> extension = Variables(notenvironment="bla", default='default')
        >>> (DictMentor().bind(extension).load_yaml(jstr) ==
        ...     {'a': 1, 'file_path': 'my_file.default.cfg'})
        True
        >>> extension = Variables(notenvironment="bla", fail_on_unset=True)
        >>> DictMentor().bind(extension).load_yaml(jstr)
        Traceback (most recent call last):
        ...
        dictmentor.extensions.ExtensionError: Variable 'environment' is unset.
    """
    __pattern__ = '.*({{var::(.*)}}).*'

    def __init__(self, fail_on_unset: bool = False, default: str = 'none', **_vars: Any):
        """
        Initializer.

        Args:
            fail_on_unset (bool): If set to True an exception will be raised when the environment
                variable is unset; otherwise the default value (see next) will be used instead.
            default (str): If a environment variable is unset, it will get this value instead.
        """
        self.fail_on_unset = bool(fail_on_unset)
        self.default = str(default)
        self.vars = _vars

    def _config(self) -> ExtensionConfig:
        """
        Tells the processor to scan for the specified pattern in node keys and values.
        Returns: The scanner configuration.
        """
        # tells the scanner to look for the specified pattern in key and value strings
        return dict(pattern=self.__pattern__, search_in_keys=True, search_in_values=True)

    def _apply(self, ctx: ExtensionContext) -> AugmentedDict:
        """
        Replaces any {{var::*}} directives with it's actual variable value or a default.

        Args:
            ctx: The processing context.

        Returns:
            Returns the altered node key and value.
        """
        node_key, node_value = ctx.node

        def process(pattern: Pattern[str], _str: str) -> Any:
            _match = pattern.match(_str)
            if _match is None:
                return _str
            # We got a match
            # Group 0: Whole match; Group 1: Our placeholder; Group 2: The environment variable
            placeholder, varname = _match.group(1), _match.group(2)
            varval = self.vars.get(varname, None)
            if varval is None and self.fail_on_unset:
                raise ExtensionError("Variable '{}' is unset.".format(varname))

            return _str.replace(placeholder, varval or self.default)

        _pattern = re.compile(self.__pattern__)
        node_key = process(_pattern, node_key)
        node_value = process(_pattern, node_value)

        return {node_key: node_value}



"""Package dictmentor."""

from . import extensions
from .base import DictMentor
from .types import Version

DictMentor = Mentor = DictMentor  # type: ignore
ext = extensions  # pylint: disable=invalid-name


def version() -> Version:
    """Return the version of the package."""
    return '0.2.1'



"""Contains custom datatypes."""

from typing import Union, IO, Dict, Any, Tuple

AugmentedDict = Dict[Any, Any]


NonAugmentedDict = Dict[Any, Any]


ExtensionConfig = Dict[str, Any]


NodeKeyVal = Tuple[str, Any]


Version = str


# A valid yaml might be
# 1. a str containing yaml data
# 2. a filepath pointing to a yaml file
# 3. a stream containing yaml data
YamlDocument = Union[str, IO[bytes]]



"""Contains core classes."""
from typing import List, Iterable, cast, Optional, Any

from ruamel import yaml  # type: ignore

from .extensions import Extension, ExtensionContext
from .types import YamlDocument, NonAugmentedDict, AugmentedDict
from .utils import dict_find_pattern
from .validator import Validator


class DictMentor:
    """
    Augments a given dictionary by applying so called extensions.
    """

    def __init__(self, *extensions: Extension):
        self._extensions = []  # type: List[Extension]
        self._init_extensions([ext for ext in extensions])

    def _init_extensions(self, extensions: Iterable[Extension]) -> None:
        if extensions is None or not extensions:
            return

        if not Validator.is_real_iterable(extensions=extensions):
            extensions = [cast(Extension, extensions)]

        for ext in extensions:
            self.bind(ext)

    def bind(self, extension: Extension) -> 'DictMentor':
        """
        Add any predefined or custom extension.

        Args:
            extension: Extension to add to the processor.

        Returns:
            The DictMentor itself for chaining.
        """
        if not Extension.is_valid_extension(extension):
            raise ValueError("Cannot bind extension due to missing interface requirements")
        self._extensions.append(extension)

        return self

    def augment(self, dct: NonAugmentedDict,
                document: Optional[YamlDocument] = None) -> AugmentedDict:
        """
        Augments the given dictionary by using all the bound extensions.

        Args:
            dct: Dictionary to augment.
            document: The document the dictionary was loaded from.

        Returns:
            The augmented dictionary.
        """
        Validator.instance_of(dict, raise_ex=True, dct=dct)

        # Apply any configured loader
        for instance in self._extensions:
            nodes = list(dict_find_pattern(dct, **instance.config()))
            for parent, k, val in nodes:
                parent.pop(k)
                fragment = instance.apply(
                    ExtensionContext(
                        mentor=self,
                        document=document or dct,
                        dct=dct,
                        parent_node=parent,
                        node=(k, val)
                    )
                )
                if fragment is not None:
                    parent.update(fragment)

        return dct

    @classmethod
    def _load_plain_yaml(cls, _yaml: YamlDocument) -> Any:
        """
        Will just load the yaml without executing any extensions. You will get the plain dictionary
        without augmentation. It is equivalent to just perform `yaml.safe_load`. Besides that you
        can specify a stream, a file or just a string that contains yaml/json data.

        Examples:
            >>> jstr = '{"a":1, "b": {"c": 3, "d": "d"}}'
            >>> d = DictMentor._load_plain_yaml(jstr)
            >>> d['a'], d['b']['c'], d['b']['d']
            (1, 3, 'd')

        Args:
            yaml_: Whether a stream (e.g. file pointer), a file name of an existing file or string
                containing yaml/json data.

        Returns:
            Returns the yaml_ data as a python dictionary.
        """
        if Validator.is_stream(yaml_=_yaml):
            return yaml.safe_load(_yaml)
        if Validator.is_file(yaml_=_yaml):
            with open(_yaml) as fhandle:  # type: ignore
                return yaml.safe_load(fhandle)
        if Validator.instance_of(target_type=str, yaml_=_yaml):
            return yaml.safe_load(_yaml)

        raise TypeError("Argument '_yaml' is whether a stream, nor a file, nor a string")

    def load_yaml(self, _yaml: YamlDocument) -> AugmentedDict:
        """
        Loads a partial yaml and augments it. A partial yaml in this context is a yaml that is
        syntactically correct, but is not yet complete in terms of content.
        The yaml will be completed by augmenting with some external resources and/or executing so
        called extensions.

        Args:
            _yaml: Whether a stream (e.g. file pointer), a file name of an existing file or string
                containing yaml data.

        Returns:
            Returns the yaml data as an augmented python dictionary.
        """
        return self.augment(self._load_plain_yaml(_yaml), document=_yaml)



from setuptools import setup, find_packages


VERSION = '0.2.1'


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='dictmentor',
    version=VERSION,
    description="A python dictionary augmentation utility",
    long_description=readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/HazardDede/dictmentor',
    author='d.muth',
    author_email='d.muth@gmx.net',
    license='MIT',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Indicate what this project is about
        'Topic :: Software Development :: Libraries',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='dict dictionary augmentation augment style',
    project_urls={
        'Documentation': 'https://github.com/HazardDede/dictmentor/blob/master/README.md',
        'Source': 'https://github.com/HazardDede/dictmentor/',
        'Tracker': 'https://github.com/HazardDede/dictmentor/issues',
    },
    packages=find_packages(exclude=["tests", "docs"]),
    install_requires=[
        'attrs>=19.0.0',
        'ruamel.yaml>=0.15.0'
    ],
    python_requires='>=3.4',
    include_package_data=True
)



"""Contains argument validating stuff."""

from functools import partial
from pathlib import Path
from typing import Dict, Any, Callable, Iterable, cast, Union, Optional

ValidationReturn = Union[bool, Dict[str, bool]]


class Validator:
    """
    Provides utility methods to easily validate arguments / variables.
    Per default the methods will return True / False whether the argument obeys the rules or not.
    If raise_ex is set to True instead of returning True / False a meaningful exception will be
    thrown.

    The advantage of this component is to produce consistent validation and error messages in a more
    convenient way that assert or other mechanisms do provide.

    But keep in mind: The most pythonic idiom is to clearly document what the function expects and
    then just try to use whatever gets passed to your function. The validators are only for
    situations were you cannot avoid it.
    """
    @staticmethod
    def __logical_dict_merge(first: Dict[Any, bool], second: Dict[Any, bool]) -> Dict[Any, bool]:
        if second is None or not isinstance(second, dict):
            return first

        res = first.copy()
        for k, val in second.items():
            if k in first:
                res[k] = res[k] and val
            else:
                res[k] = val

        return res

    @staticmethod
    def __test_all(condition: Callable[[str, Any], bool],
                   formatter: Callable[[str, Any], str], raise_ex: bool, summary: bool,
                   validators: Optional[Iterable[Callable[..., ValidationReturn]]] = None,
                   **items: Any) -> ValidationReturn:
        assert callable(condition)
        assert callable(formatter)

        res = {}  # type: Dict[Any, bool]
        if validators is not None:
            for validator in validators:
                # Case 1: raise_ex=True -> We don't have anything to catch
                #   -> just propagate the error
                # Case 2: summary=True
                #   -> We have to short circuit if the result is evaluated to False
                # Case 3: raise_ex=False, summary=False
                #   -> We get a dictionary containing the evaluated conditions
                # -> We need to merge them with the previous output / results
                val_res = validator(raise_ex=raise_ex, summary=summary, **items)
                if summary and not val_res:
                    return False
                res = Validator.__logical_dict_merge(res, cast(Dict[Any, bool], val_res))

        cond_res = {}
        for varname, val in items.items():
            cond_eval = condition(varname, val)
            if raise_ex and not cond_eval:
                # Short-circuit on raise_ex
                raise ValueError(formatter(varname, val))
            if summary and not cond_eval:
                # Short-circuit on summary (at least one item has validated to False -> abort)
                return False
            cond_res[varname] = cond_eval

        res = Validator.__logical_dict_merge(res, cond_res)
        # If summary is True we need to return back only one boolean value.
        return res if not summary else all(res.items())

    @staticmethod
    def instance_of(target_type: Optional[type] = None, raise_ex: bool = False,
                    summary: bool = True, **items: Any) -> ValidationReturn:
        """
        Tests if the given key-value pairs (items) are instances of the given ``target_type``.
        Per default this function yields whether ``True`` or ``False`` depending on the fact if all
        items withstand the validation or not. Per default the validation / evaluation is
        short-circuit and will return as soon an item evaluates to ``False``.

        When ``raise_ex`` is set to ``True`` the function will raise a meaningful error message
        after the first item evaluates to ``False`` (short-circuit).

        When ``summary`` is set to ``False`` a dictionary is returned containing the individual
        evaluation result of each item (non short-circuit).

        Examples:


            >>> Validator.instance_of(my_str='str', my_str2='str2', target_type=str)
            True
            >>> Validator.instance_of(my_str1='str', target_type=int)
            False
            >>> Validator.instance_of(my_str=None, raise_ex=True, target_type=str)
            Traceback (most recent call last):
                ...
            ValueError: 'my_str' (NoneType) is not an instance of type 'str'
            >>> Validator.instance_of(my_str='a', my_str2=1, raise_ex=True, target_type=str)
            Traceback (most recent call last):
                ...
            ValueError: 'my_str2' (int) is not an instance of type 'str'
            >>> (Validator.instance_of(
            ...     my_str='str', my_str2='str2', non_str=5, target_type=str, summary=False
            ... )
            ... == {'my_str': True, 'my_str2': True, 'non_str': False})
            True

        Args:
            raise_ex (bool, optional): If set to ``True`` an exception is raised if at least one
            item is validated to ``False`` (works short-circuit and will abort the validation when
            the first item is evaluated to ``False``).

            summary (bool, optional): If set to ``False`` instead of returning just a single
            `bool` the validation will return a dictionary containing the individual evaluation
            result of each item.

            target_type (type): The target type to test the values against.

        Returns:
            (boolean or dictionary): ``True`` when the value was successfully validated; ``False``
                otherwise.
            If ``summary`` is set to ``False`` a dictionary containing the individual evaluation
            result of each item will be returned.
            If ``raise_ex`` is set to True, instead of returning False a meaningful error will be
            raised.
        """
        if not target_type:
            raise ValueError("Argument 'target_type' is None")

        return Validator.__test_all(
            condition=lambda _, val: isinstance(val, cast(type, target_type)),
            formatter=(
                lambda name, val:
                "'{varname}' ({actual}) is not an instance of type '{ttype}'".format(
                    varname=name,
                    actual=type(val).__name__,
                    ttype=cast(type, target_type).__name__
                )
            ),
            raise_ex=raise_ex,
            summary=summary,
            **items
        )

    @staticmethod
    def is_stream(raise_ex: bool = False, summary: bool = True,
                  **items: Any) -> ValidationReturn:
        """
        Tests if the given key-value pairs (items) are all streams. Basically checks if the item
        provides a read(...) method.
        Per default this function yields whether ``True`` or ``False`` depending on the fact if all
        items withstand the validation or not. Per default the validation / evaluation is
        short-circuit and will return as soon an item evaluates to ``False``.

        When ``raise_ex`` is set to ``True`` the function will raise a meaningful error message
        after the first item evaluates to ``False`` (short-circuit).

        When ``summary`` is set to ``False`` a dictionary is returned containing the individual
        evaluation result of each item (non short-circuit).

        Examples:

            >>> class MyStream():
            ...     def read(self):
            ...         pass
            >>> Validator.is_stream(stream=MyStream())
            True
            >>> Validator.is_stream(nonstream='abc')
            False
            >>> Validator.is_stream(stream=MyStream(), nonstream='abc', summary=True)
            False
            >>> (Validator.is_stream(stream=MyStream(), nonstream='abc', summary=False)
            ... == {'stream': True, 'nonstream': False})
            True
            >>> Validator.is_stream(nonstream='abc', raise_ex=True)
            Traceback (most recent call last):
                ...
            ValueError: 'nonstream' is not a stream

        Args:

            raise_ex (bool, optional): If set to ``True`` an exception is raised if at least one
            item is validated to ``False`` (works short-circuit and will abort the validation when
            the first item is evaluated to ``False``).

            summary (bool, optional): If set to ``False`` instead of returning just a single
            ``bool`` the validation will return a dictionary containing the individual evaluation
            result of each item.

        Returns:
            (boolean or dictionary): ``True`` when the value was successfully validated; ``False``
                otherwise.
            If ``summary`` is set to ``False`` a dictionary containing the individual evaluation
            result of each item will be returned.
            If ``raise_ex`` is set to True, instead of returning False a meaningful error will be
            raised.

        """
        return Validator.__test_all(
            condition=lambda _, val: getattr(val, 'read', None) is not None,
            formatter=lambda name, _: "'{varname}' is not a stream".format(varname=name),
            raise_ex=raise_ex,
            summary=summary,
            **items
        )

    @staticmethod
    def is_file(raise_ex: bool = False, summary: bool = True,
                **items: Any) -> ValidationReturn:
        """
        Tests if the given key-value pairs (items) are physical existent files or links to regular
        files.
        Per default this function yields whether ``True`` or ``False`` depending on the fact if all
        items withstand the validation or not. Per default the validation / evaluation is
        short-circuit and will return as soon an item evaluates to ``False``.

        When ``raise_ex`` is set to ``True`` the function will raise a meaningful error message
        after the first item evaluates to ``False`` (short-circuit).

        When ``summary`` is set to ``False`` a dictionary is returned containing the individual
        evaluation result ofeach item (non short-circuit).

        Examples:

            >>> import tempfile
            >>> tmp = tempfile.NamedTemporaryFile()
            >>> Validator.is_file(file=tmp.name)
            True
            >>> Validator.is_file(nonfile="iamsurethatthisdontexist")
            False
            >>> Validator.is_file(file=tmp.name, nonfile="iamsurethatthisdontexist")
            False
            >>> (Validator.is_file(file=tmp.name, nonfile="iamsurethatthisdontexist", summary=False)
            ... == {'file': True, 'nonfile': False})
            True
            >>> Validator.is_file(file=tmp.name, nonfile="iamsurethatthisdontexist", raise_ex=True)
            Traceback (most recent call last):
                ...
            ValueError: 'nonfile' is not a file
            >>> with open(tmp.name) as fp:
            ...     Validator.is_file(nonfile=fp)
            False


        Args:

            raise_ex (bool, optional): If set to ``True`` an exception is raised if at least one i
            tem is validated to ``False`` (works short-circuit and will abort the validation when
            the first item is evaluated to ``False``).

            summary (bool, optional): If set to ``False`` instead of returning just a single
            ``bool`` the validation will return a dictionary containing the individual evaluation
            result of each item.

        Returns:
            (boolean or dictionary): ``True`` when the value was successfully validated; ``False``
                otherwise.
            If ``summary`` is set to ``False`` a dictionary containing the individual evaluation
            result of each item will be returned.
            If ``raise_ex`` is set to True, instead of returning False a meaningful error will be
            raised.

        """
        return Validator.__test_all(
            condition=lambda _, val: Path(val).is_file(),
            formatter=lambda name, _: "'{varname}' is not a file".format(varname=name),
            validators=[partial(Validator.instance_of, target_type=str)],
            raise_ex=raise_ex,
            summary=summary,
            **items
        )

    @staticmethod
    def is_real_iterable(raise_ex: bool = False, summary: bool = True,
                         **items: Any) -> ValidationReturn:
        """
        Tests if the given items are iterables (collections) but no strings.
        Per default this function yields whether ``True`` or ``False`` depending on the fact if all
        items withstand the validation or not. Per default the validation / evaluation is
        short-circuit and will return as soon an item evaluates to ``False``.

        When ``raise_ex`` is set to ``True`` the function will raise a meaningful error message
        after the first item evaluates to ``False`` (short-circuit).

        When ``summary`` is set to ``False`` a dictionary is returned containing the individual
        evaluation result of each item (non short-circuit).

        Examples:

            >>> l = ['i', 'am', 'an', 'iterable']
            >>> Validator.is_real_iterable(l=l)
            True

            >>> d = {'i': 'am', 'a': 'dict'}
            >>> Validator.is_real_iterable(d=d)
            True

            >>> s = "i am a string"
            >>> Validator.is_real_iterable(s=s)
            False

            >>> Validator.is_real_iterable(raise_ex=True, s=s)
            Traceback (most recent call last):
                ...
            ValueError: 's' is not an iterable

        Args:

            raise_ex (bool, optional): If set to ``True`` an exception is raised if at least one
            item is validated to ``False`` (works short-circuit and will abort the validation when
            the first item is evaluated to ``False``).

            summary (bool, optional): If set to ``False`` instead of returning just a single
            ``bool`` the validation will return a dictionary containing the individual evaluation
            result of each item.

        Returns:
            (boolean or dictionary): ``True`` when the value was successfully validated; ``False``
                otherwise.
            If ``summary`` is set to ``False`` a dictionary containing the individual evaluation
            result of each item will be returned.
            If ``raise_ex`` is set to True, instead of returning False a meaningful error will be
            raised.

        """

        return Validator.__test_all(
            condition=lambda _, val: hasattr(val, '__iter__') and not isinstance(val, str),
            formatter=lambda name, _: "'{varname}' is not an iterable".format(varname=name),
            raise_ex=raise_ex,
            summary=summary,
            **items
        )
